"""
A collection of pure functions used to calculate wing properties.
"""

import jax
import jax.numpy as jnp

@jax.jit
def wing_area_ar(thetas, b, c):
    z = -(b / 2) * jnp.cos(thetas)
    area = jnp.trapezoid(y=c, x=z)
    ar = b ** 2 / area

    return jnp.array([area, ar])

@jax.jit
def cl(coefficients, aspect_ratio):
    return coefficients[0] * jnp.pi * aspect_ratio

@jax.jit
def cdi(cl, efficiency, aspect_ratio):
    return cl ** 2 / (efficiency * jnp.pi * aspect_ratio)

@jax.jit
def efficiency(n_list, coefficients):
    a1 = coefficients[0]

    delta = jnp.einsum("i,i->", n_list[1:], (coefficients[1:] / a1) ** 2)

    return 1 / (1 + delta)

@jax.jit
def lift(cl, wing_area, rho, v_infty):
    return 0.5 * cl * wing_area * rho * v_infty ** 2

@jax.jit
def induced_drag(cdi, wing_area, rho, v_infty):
    return 0.5 * cdi * wing_area * rho * v_infty ** 2

@jax.jit
def unitless_circulation(n_list, thetas, coefficients):
    gamma = jnp.sin(jnp.einsum("i,j->ij", thetas, n_list))  # Shape (N, F)
    gamma = jnp.einsum("j,ij->i", coefficients, gamma)  # Shape (N,)

    return gamma

@jax.jit
def alpha_induced(n_list, thetas, coefficients):
    terms_1 = jnp.einsum("i,j,j->ij", 1 / jnp.sin(thetas), coefficients, n_list)  # Shape (N, F)
    terms_2 = jnp.sin(jnp.einsum("i,j->ij", thetas, n_list))  # Shape (N, F)

    a_i = jnp.einsum("ij,ij->i", terms_1, terms_2)  # Shape (N,)

    return a_i


from typing import Dict, Literal, Any, List

import jax
import jax.numpy as jnp
from functools import cached_property
import warnings
from .wing import UnsweptWing
from .util import aerodynamic_calculator as calc
from src.lazyllt.util.fourier_solver import a_i_rms_error

def requires_gradient(func):
    def wrapper(self, *args, **kwargs):
        if self.gradients is None:
            raise ValueError("Called a method which requires gradients, but the gradients were not loaded. Try setting calculate_gradients=True in LiftingLineModel.solve.")
        else:
            return func(self, *args, **kwargs)

    return wrapper

class LiftingLineSolution:
    wing: UnsweptWing

    def __init__(self, wing: UnsweptWing, n_list, coefficients, gradients):
        """
        Solution class for a lifting-line calculation. This class should only be initialized by LiftingLineModel, never
        by the user.
        :param wing: The wing the solution was calculated for.
        :param n_list: List of Fourier coefficient indices. Shape (F,)
        :param coefficients: Fourier coefficients. Shape (F,)
        :param gradients: A dictionary holding named parameters and the
        gradient of the Fourier coefficients wrt the parameter, held in an array of shape (F,).
        """
        self.wing = wing
        self.n_list = n_list
        self.coefficients = coefficients
        self.gradients = gradients

        if gradients:
            #Mapping of strings to properties for gradient information.
            #callable is the jacobian function.
            #parameters is the values that the function is differentiated wrt
            #dependencies is the gradients that need to be calculated (within function_mappings) first
            self.function_mappings: Dict[str, Dict[Literal["callable", "parameters", "dependencies"], Any]] = {
                "cl": {
                    "callable": lambda: jax.jacrev(calc.cl, argnums=(0, 1))(self.coefficients, self.wing.aspect_ratio),
                    "parameters": ["coefficients", "aspect_ratio"],
                    "dependencies": []
                },
                "efficiency": {
                    "callable": lambda: jax.jacrev(calc.efficiency, argnums=(1,))(self.n_list, self.coefficients),
                    "parameters": ["coefficients"],
                    "dependencies": []
                },
                "cdi": {
                    "callable": lambda: jax.jacrev(calc.cdi, argnums=(0, 1, 2))(self.cl, self.efficiency, self.wing.aspect_ratio),
                    "parameters": ["cl", "efficiency", "aspect_ratio"],
                    "dependencies": ["cl", "efficiency"]
                },
                "lift": {
                    "callable": lambda: jax.jacrev(calc.lift, argnums=(0, 1))(self.cl, self.wing.wing_area, 1.0, 1.0),
                    "parameters": ["cl", "area"],
                    "dependencies": ["cl"]
                },
                "induced_drag": {
                    "callable": lambda: jax.jacrev(calc.induced_drag, argnums=(0, 1))(self.cdi, self.wing.wing_area, 1.0, 1.0),
                    "parameters": ["cdi", "area"],
                    "dependencies": ["cdi"]
                }
            }

        error = self.solver_error
        if error > 0.1:
            warnings.warn(f"The solver had an induced angle-or-attack RMS error of {error:.5f} degrees, "
                          "indicating possible numerical issues. You may have an excessively high value for "
                          "num_points and/or num_coefficients.")

    @cached_property
    def solver_error(self):
        """
        The RMS error in calculating the induced angle of attack in degrees.
        A useful metric for detecting silent numerical issues.
        Note that this metric shows how well the solver fit the solution given the provided constraints,
        not the accuracy of the aerodynamic properties.
        """
        return jnp.rad2deg(a_i_rms_error(
            self.n_list,
            self.coefficients,
            self.wing.thetas,
            self.wing.c,
            self.wing.alpha_geo,
            self.wing.b,
            self.wing.alpha_0
        ))

    @cached_property
    @requires_gradient
    def design_variables(self):
        """
        A list of design variables that a wing property can be differentiated with respect to.
        Requires gradients to be loaded.
        """
        return set(self.gradients["coefficients"].keys())

    @cached_property
    @requires_gradient
    def differentiable_properties(self):
        """
        A list of differentiable properties of the wing.
        Requires gradients to be loaded.
        """
        #Both loaded and unloaded properties
        return set(self.function_mappings.keys() | self.gradients.keys())

    @requires_gradient
    def calculate_gradients(self, properties: List[str], design_variables: List[str]):
        """
        Calculates the gradients of a list of wing properties with respect to a list of design variables.
        Populates the calculated gradients in the ``gradients`` attribute.
        Requires gradients to be loaded. Note that all gradients are calculated in MKS+degrees.
        :param properties: A list of wing properties
        :param design_variables: A list of design variables
        :return: The dictionary of all calculated gradients.
        """

        for prop in properties:
            if prop not in self.differentiable_properties:
                warnings.warn(f"{prop} is not a differentiable wing property; skipped.")
                continue

            # Create dictionary for derivatives if not existing
            self.gradients[prop] = self.gradients.get(prop, dict())

            mapping = self.function_mappings[prop]

            #Calculate the partials of all the dependency functions
            dependencies = mapping["dependencies"]
            if len(dependencies) > 0:
                self.calculate_gradients(dependencies, design_variables)

            #Calculate the jacobian wrt the parameters of the function and populate the gradients dict
            jacobian = mapping["callable"]()
            for param, grad in zip(mapping["parameters"], jacobian):
                self.gradients[prop][param] = grad

            for dv in design_variables:
                if dv not in self.design_variables:
                    warnings.warn(f"{dv} is not a valid design variable; skipped.")
                    continue

                # Do not recalculate if the gradient is already there
                if self.gradients[prop].get(dv) is not None:
                    continue

                total_grad = jnp.float32(0)
                for param in mapping["parameters"]:
                    #Only calculate the partial if it exists
                    base_grad = self.gradients[param].get(dv)
                    if base_grad is not None:
                        total_grad += jnp.dot(base_grad, self.gradients[prop][param])
                self.gradients[prop][dv] = total_grad

        return self.gradients

    @cached_property
    def cl(self):
        """
        Wing coefficient of lift.
        """
        return calc.cl(self.coefficients, self.wing.aspect_ratio)


    @cached_property
    @requires_gradient
    def cl_slope(self):
        """
        Helper attribute to get slope of the wing coefficient of lift w.r.t. angle of attack in radians.
        """
        self.calculate_gradients(["cl"], ["aoa"])

        #Since the slope gradient is in units deg^-1, it must be divided by the deg2rad factor
        return jnp.rad2deg(self.gradients["cl"]["aoa"])

    @cached_property
    def cdi(self):
        """
        Coefficient of induced drag.
        """
        return calc.cdi(self.cl, self.efficiency, self.wing.aspect_ratio)

    @cached_property
    def efficiency(self):
        """
        Oswald efficiency factor.
        """
        return calc.efficiency(self.n_list, self.coefficients)

    def lift(self, rho, v_infty):
        """
        :param rho: Air density in kg/m^3.
        :param v_infty: Freestream velocity in m/s.
        :return: The lift force in newtons.
        """
        return calc.lift(self.cl, self.wing.wing_area, rho, v_infty)

    def induced_drag(self, rho, v_infty):
        """
        :param rho: Air density in kg/m^3.
        :param v_infty: Freestream velocity in m/s.
        :return: The induced drag force in newtons.
        """
        return calc.induced_drag(self.cdi, self.wing.wing_area, rho, v_infty)

    def circulation(self, v_infty):
        """
        Gets the circulation distribution of the wing. The distribution is evenly spaced.
        :param v_infty: The freestream velocity in m/s.
        :return: Circulation distribution in m^2/s. Shape (N,)
        """

        return 2 * self.wing.b * v_infty * calc.unitless_circulation(self.n_list, self.wing.thetas, self.coefficients)

    @cached_property
    def alpha_induced(self):
        """
        The induced angle of attack distribution in degrees. Shape (N,)
        """
        return jnp.rad2deg(calc.alpha_induced(self.n_list, self.wing.thetas, self.coefficients))

    @cached_property
    def alpha_effective(self):
        """
        The effective angle of attack distribution in degrees.
        """
        return jnp.rad2deg(self.wing.alpha_geo) - self.alpha_induced #Shape (N,)


import unittest
from src.lazyllt import LiftingLineModel, UnsweptWing
import jax.numpy as jnp


# Test that the solver works with an elliptical distribution.
class TestElliptical(unittest.TestCase):
    #Make sure that the Oswald efficiency factor for an elliptically loaded wing is close to 1.
    def test_efficiency(self):
        model = LiftingLineModel()
        wing = UnsweptWing(
            span=5,
            root_chord=1,
            alpha_0=-1
        )
        wing.set_taper_function(lambda z, _: jnp.sqrt(1 - z ** 2))
        wing.parametric_taper()

        model.add_wing(wing)
        solution = next(model.solve())

        self.assertAlmostEqual(solution.efficiency, 1, places=4)

    #Test that the lift slope is theoretically correct for an elliptically loaded wing.
    def test_lift_slope(self):
        model = LiftingLineModel()
        wing = UnsweptWing(
            span=5,
            root_chord=1,
            alpha_0=-1
        )
        wing.set_taper_function(lambda z, _: jnp.sqrt(1 - z ** 2))
        wing.parametric_taper()

        model.add_wing(wing)
        solution = next(model.solve(calculate_gradients=True))
        slope = solution.cl_slope
        theoretical_slope = 2 * jnp.pi / (1 + 2 / wing.aspect_ratio)

        self.assertAlmostEqual(slope, theoretical_slope, places=2)

    #Test that the integrated circulation matches the total lift within 1%
    def test_circulation(self):
        model = LiftingLineModel()
        wing = UnsweptWing(
            span=5,
            root_chord=1,
            alpha_0=-1
        )
        wing.set_taper_function(lambda z, _: jnp.sqrt(1 - z ** 2))
        wing.parametric_taper()

        model.add_wing(wing)
        solution = next(model.solve(calculate_gradients=True))

        v_infty = 100
        rho = 1
        l_prime = rho * v_infty * solution.circulation(v_infty)
        lift = jnp.trapezoid(l_prime, wing.nodes).item()

        self.assertAlmostEqual(lift, solution.lift(rho, v_infty), delta=lift/100)

if __name__ == "__main__":
    unittest.main()


import warnings
from typing import Callable, Dict, Optional

import jax
import jax.numpy as jnp
from jaxtyping import Array
from .util import aerodynamic_calculator as calc


class UnsweptWing:
    #Scalar properties
    num_points: int
    b: float
    alpha_0: float
    root_chord: float

    #Array properties
    thetas: Array
    c: Array
    alpha_geo: Array

    taper_fn: Callable[[Array, Dict], Array]
    taper_params: Dict
    twist_fn: Callable[[Array, Dict], Array]
    twist_params: Dict

    #Computed properties
    @property
    def wing_area(self):
        return calc.wing_area_ar(self.thetas, self.b, self.c)[0]

    @property
    def aspect_ratio(self):
        return calc.wing_area_ar(self.thetas, self.b, self.c)[1]

    @property
    def nodes(self):
        return -0.5 * self.b * jnp.cos(self.thetas)

    def __init__(self, span: float, root_chord: float, alpha_0: float, aoa: float=0.0, num_points: int = 50):
        """
        A wing without sweep or dihedral.
        By default, a rectangular wing with no twist.
        :param span: The wingspan in meters.
        :param root_chord: The length of the chord at the wing root, in meters.
        :param alpha_0: The zero-lift angle of the wing, in degrees. This is a property of the airfoil geometry.
        :param aoa: The angle of attack of the wing, in degrees.
        :param num_points: The number of points used to discretize the wing. Must be at least 4.
        """
        if num_points < 4:
            raise ValueError(f"num_points must be at least 4; got {num_points}.")
        if span <= 0:
            raise ValueError(f"span must be greater than 0; got {span}.")
        if root_chord <= 0:
            raise ValueError(f"root_chord must be greater than 0; got {root_chord}.")
        if alpha_0 > 0:
            warnings.warn(f"Expected a nonpositive value for alpha_0; got {alpha_0}. This may be a mistake, because "
                          "it implies the airfoil generates negative lift at zero angle of attack.")
        if num_points > 1000:
            warnings.warn(f"num_points values of over 1000 can cause silent numerical issues; got {num_points}.")

        self.num_points = num_points
        self.thetas = jnp.linspace(1e-3, jnp.pi - 1e-3, num_points) #Shape (N,)

        self.root_chord = jnp.float32(root_chord)
        self.c = jnp.full((num_points,), self.root_chord) #Shape (N,)
        self.b = jnp.float32(span)

        self.aoa = jnp.deg2rad(aoa)
        self.alpha_geo = jnp.zeros_like(self.thetas) + self.aoa #Shape (N,)
        self.alpha_0 = jnp.deg2rad(alpha_0).item()

        self.taper_fn = lambda z, t: self.c
        self.taper_params = dict()
        self.twist_fn = lambda z, t: self.alpha_geo
        self.twist_params = dict()

    def __str__(self):
        return f"""
Wing {{
    span = {self.b} m
    root_chord = {self.root_chord} m
    alpha_0 = {jnp.rad2deg(self.alpha_0).item():.5f} deg
    aoa = {jnp.rad2deg(self.aoa).item():.5f} deg
    taper_params = {self.taper_params}
    twist_params = {self.twist_params}
    num_points = {self.num_points}
}}"""

    def __repr__(self):
        return self.__str__()

    def set_taper_function(self, f: Callable[[Array, Dict[str, float]], Array]):
        """
        Set a custom parametric chord distribution function.
        :param f: A function(z, t)->c such that 0 <= z_i <= 1 and 0 < c_i <= 1.
            z_i=0 is the wing base and z_i=1 is either wing tip. t is a dictionary of named parameters.
            c_i is a fraction of the root chord length.
        :return:
        """
        self.taper_fn = f

    def parametric_taper(self, parameters: Optional[Dict[str, float]] = None):
        """
        Evaluate the currently set parametric chord distribution function.
        :param parameters: A dictionary of named parameters.
        :return:
        """
        if parameters is None:
            parameters = dict()

        # Make z go from 1 to 0 to 1 for easier scaling
        z = jnp.abs(-jnp.cos(self.thetas))

        self.taper_params = parameters
        self.c = jnp.float32(self.root_chord) * self.taper_fn(z, parameters)

    def linear_taper(self, taper_ratio):
        """
        Sets a linearly tapered chord distribution.
        :param taper_ratio: The ratio between the length of the tip chord and root chord. Must be between 0 and 1.
        :return:
        """
        if taper_ratio <= 0 or taper_ratio > 1:
            raise ValueError(f"taper_ratio must be between 0 and 1; got {taper_ratio}.")


        def f(z, params):
            ratio = params["taper_ratio"]
            c = 1 - (1 - ratio) * z

            return c

        self.set_taper_function(f)
        self.parametric_taper({"taper_ratio": jnp.float32(taper_ratio)})

    def taper_gradient(self):
        """
        Returns the partial derivative of the wing chord distribution with respect to the taper parameters.
        :return: A dictionary mapping named parameters to a gradient array of shape (N,).
        """
        # Make z go from 1 to 0 to 1 for easier scaling
        z = jnp.abs(-jnp.cos(self.thetas))

        if self.taper_params and len(self.taper_params) > 0:
            gradients: Dict = jax.jacrev(self.taper_fn, argnums=1)(z, self.taper_params)
        else:
            gradients = dict()

        #The gradients need to be scaled based on the root chord
        for param in gradients.keys():
            gradients[param] *= self.root_chord

        #Since the root chord is a scaling variable, its gradient is just the unscaled chord distribution
        gradients["root_chord"] = self.taper_fn(z, self.taper_params)

        return gradients

    def set_twist_function(self, f: Callable[[Array, Dict[str, float]], Array]):
        """
        Set a custom parametric chord distribution function.
        :param f: A function(z, t)->a such that 0 <= z_i <= 1.
            z_i=0 is the wing base and z_i=1 is either wing tip. t is a dictionary of named parameters.
            a_i is the twist angle at a point in degrees.
        :return:
        """
        self.twist_fn = f

    def parametric_twist(self, parameters: Optional[Dict[str, float]] = None):
        """
        Evaluate the currently set parametric twist distribution function.
        :param parameters: A dictionary of named parameters.
        :return:
        """
        if parameters is None:
            parameters = dict()

        # Make z go from 1 to 0 to 1 for easier scaling
        z = jnp.abs(-jnp.cos(self.thetas))

        self.twist_params = parameters
        self.alpha_geo = jnp.deg2rad(self.twist_fn(z, parameters)) + self.aoa

    def linear_twist(self, root_twist: float, tip_twist: float):
        """
        Sets a linear twist distribution.
        :param root_twist: Twist at the wing root, in degrees.
        :param tip_twist: Twist at the wing tip, in degrees.
        :return:
        """

        def f(z, params):
            root = params["root_twist"]
            tip = params["tip_twist"]

            return root_twist + (root - tip) * z


        self.set_twist_function(f)
        self.parametric_twist({
            "root_twist": jnp.deg2rad(root_twist).item(),
            "tip_twist": jnp.deg2rad(tip_twist).item()
        })


    def twist_gradient(self):
        """
        Returns the partial derivative of the wing twist distribution with respect to the twist parameters.
        :return: A dictionary mapping named parameters to a gradient array of shape (N,).
        """
        # Make z go from 1 to 0 to 1 for easier scaling
        z = jnp.abs(-jnp.cos(self.thetas))

        if self.twist_params and len(self.twist_params) > 0:
            gradients: Dict = jax.jacrev(self.twist_fn, argnums=1)(z, self.twist_params)
        else:
            gradients = dict()

        #Since the angle of attack just adds on to the current angle, its gradient is just 1 degree per degree.
        gradients["aoa"] = jnp.ones_like(z)

        return gradients



import jax
import jax.numpy as jnp

@jax.jit
def solve_coefficients(n_list, thetas, c, alpha_geo, b, alpha_0):
    """
    Solve for the Fourier coefficients using matrix pseudo-inverses.
    :param n_list: Shape (F,)
    :param thetas: Shape (N,)
    :param c: Shape (N,)
    :param alpha_geo: Shape (N,). In radians.
    :param b: Scalar.
    :param alpha_0: Scalar. In radians.
    :return: Shape (F,)
    """
    a_i = jnp.einsum("i,j->ij", 1/jnp.sin(thetas), n_list) #Shape (N, F)
    a_i *= jnp.sin(jnp.einsum("i,j->ij", thetas, n_list))

    a_eff = jnp.sin(jnp.einsum("i,j->ij", thetas, n_list)) #Shape (N, F)
    a_eff = jnp.einsum("i,ij->ij", 1/c, a_eff) #Shape (N, F)
    a_eff *= 2 * b / jnp.pi

    #lhs @ coefficients = rhs
    lhs = a_i + a_eff #Shape (N, F)
    rhs = alpha_geo - alpha_0 #Shape (N,)

    pseudo_inverse = jnp.linalg.pinv(lhs) #Shape (F, N)

    coef = pseudo_inverse @ rhs # (F, N) @ (N, 1) = (F, 1)

    return coef.reshape(-1)

@jax.jit
def a_i_rms_error(n_list, coefficients, thetas, c, alpha_geo, b, alpha_0):
    """
    Get the RMS error in induced angle of attack (in radians) given calculated coefficients.
    :param n_list: Shape (F,)
    :param coefficients: Shape (F,)
    :param thetas: Shape (N,)
    :param c: Shape (N,)
    :param alpha_geo: Shape (N,). In radians.
    :param b: Scalar.
    :param alpha_0: Scalar. In radians.
    :return: Shape (F,)
    """
    a_i = jnp.einsum("i,j->ij", 1/jnp.sin(thetas), n_list) #Shape (N, F)
    a_i *= jnp.sin(jnp.einsum("i,j->ij", thetas, n_list))

    a_eff = jnp.sin(jnp.einsum("i,j->ij", thetas, n_list)) #Shape (N, F)
    a_eff = jnp.einsum("i,ij->ij", 1/c, a_eff) #Shape (N, F)
    a_eff *= 2 * b / jnp.pi

    #lhs @ coefficients = rhs
    lhs = a_i + a_eff #Shape (N, F)
    rhs = alpha_geo - alpha_0 #Shape (N,)

    residual = lhs @ coefficients - rhs #Shape (N,)
    rms_error = jnp.sqrt(jnp.mean(residual ** 2))

    return rms_error


import jax
import jax.numpy as jnp
from typing import List
from collections.abc import Generator
from jaxtyping import Float, Array

from .solution import LiftingLineSolution
from .wing import UnsweptWing
from src.lazyllt.util.fourier_solver import solve_coefficients
from .util import aerodynamic_calculator as calc


class LiftingLineModel:
    n_list: Float[Array, "F"]
    wings: List[UnsweptWing]

    def __init__(self, num_coefficients: int = 35):
        """
        A class used to manage the lifting-line problem.
        :param num_coefficients: The number of Fourier coefficients to use in the solver. Must be at least 1.
        """

        #The solver only uses the odd Fourier coefficients to ensure symmetry.
        self.n_list = jnp.arange(1, num_coefficients * 2 + 1, 2) #Shape (F,)

        self.wings = []

    def add_wing(self, wing: UnsweptWing):
        self.wings.append(wing)

    def get_wings(self) -> List[UnsweptWing]:
        return self.wings

    def solve(self, calculate_gradients=False) -> Generator[LiftingLineSolution, None, None]:
        """
        Lazily calculates the lifting-line solution for each of the given wings.
        Throws an error if no wings have been added to the model.
        :param calculate_gradients: Whether to calculate gradients or not. Disabled by default, useful for optimization.
        :return: A generator yielding LiftingLineSolution(s).
        """
        if len(self.wings) < 1:
            raise ValueError("The model does not have any loaded wings. Try calling add_wing.")

        if calculate_gradients:
            coefficients_grad = jax.jit(jax.jacrev(solve_coefficients, argnums=(2, 3, 4, 5)))
            area_ar_grad = jax.jit(jax.jacrev(calc.wing_area_ar, argnums=(1, 2)))
        else:
            coefficients_grad = lambda: None
            area_ar_grad = lambda: None

        for wing in self.wings:
            #Get the Fourier coefficients
            coefficients = solve_coefficients(self.n_list, wing.thetas, wing.c, wing.alpha_geo, wing.b, wing.alpha_0)

            if calculate_gradients:
                #Calculate gradients wrt Fourier coefficients
                c_grad, a_geo_grad, b_grad, a_0_grad = coefficients_grad(self.n_list, wing.thetas, wing.c, wing.alpha_geo, wing.b, wing.alpha_0)
                #Convert all angle gradients to degrees
                #Since all the gradients are in units rad^-1, they must be divided by the rad2deg multiplier
                a_geo_grad = jnp.deg2rad(a_geo_grad)
                a_0_grad = jnp.deg2rad(a_0_grad)
                #Calculate gradients wrt wing area and ar
                (b_grad_area, b_grad_ar), (c_grad_area, c_grad_ar) = area_ar_grad(wing.thetas, wing.b, wing.c)

                #Populate the gradients wrt scalar variables
                #The gradient dictionary is indexed first by a value, then by the variable wrt it is differentiated.
                gradients = dict()
                gradients["coefficients"] = dict()
                gradients["area"] = dict()
                gradients["aspect_ratio"] = dict()

                gradients["coefficients"]["b"] = b_grad
                gradients["area"]["b"] = b_grad_area
                gradients["aspect_ratio"]["b"] = b_grad_ar
                gradients["coefficients"]["alpha_0"] = a_0_grad

                #Calculate the parametric gradients wrt Fourier coefficients
                for parameter, param_grad in wing.taper_gradient().items():
                    #param_grad is of shape (N,).
                    #c_grad is of shape (F, N)
                    #c_grad_area and c_grad_ar are of shape (N,)
                    gradients["coefficients"][parameter] = c_grad @ param_grad #Shape (F,)
                    gradients["area"][parameter] = c_grad_area @ param_grad
                    gradients["aspect_ratio"][parameter] = c_grad_ar @ param_grad

                for parameter, param_grad in wing.twist_gradient().items():
                    # param_grad is of shape (N,).
                    # a_geo_grad is of shape (F, N)
                    gradients["coefficients"][parameter] = a_geo_grad @ param_grad  # Shape (F,)
            else:
                gradients = None

            yield LiftingLineSolution(wing, self.n_list, coefficients, gradients)